
// Simple rectifier circuit analysis (linearized diode model)
// Warning: not validated, not optimized

// Rectifier circuit:
//           _____     D
//     o----|_____|---->I---o------o
//            Rac           |
//   Uac                    |          Uout, Rz
//                         === C
//                          |
//     o--------------------o------o
//

// Method: Runge-Kutta 4
// Needs linking with math library (GCC option -lm)
// Copyright: (c) Petr Peringer, FIT TU Brno, 2018-2020

#include <stdio.h>
#include <math.h>               // sin(), M_PI

// System: rectifier1
#define SIZE 1                  // state vector size of the system
// model parameters:
const double Rac = 1;
      double Rz = 100;
const double C = 1000e-6;
// linearized diode parameters:
const double Rd = 0.2;  // [Ohm] diode resistance: linear approximation
const double Up = 0.6;  // [V] minimal diode voltage drop
      double Id;        // global variable for output only

// input:
const double Uac  = 10;         // Amplitude (NOT effective)
const double Freq = 50;         // frequency

inline double uin(double t) {
    return Uac*sin(2*M_PI*Freq*t); // sin(\omega t)
}

inline double uout(double t, double y[]) {
    double u_C = y[0];
    return u_C;
}

inline double iout(double t, double y[]) {
    return uout(t,y)/Rz;
}

// "Dynamic section" = continuous system description
// Input:  t = time of step start,
//         y = state vector = integrator outputs
// Output: yin = vector of derivatives = integrators input values
void f(double t, double *y, double *yin)
{
    double u_C   = y[0];        // u_C = uout
    double u_RD  = uin(t)-u_C;  // voltage on Rac and D
    double i_z   = u_C/Rz;      // load current
    // linearized diode model:
    if(u_RD>Up) {
        // series connection of  Rd + Rac (and source Up)
        Id = (u_RD-Up)/(Rac+Rd);
    }else{
        Id = 0;
    }
    double i_C = Id - i_z;
    yin[0] = i_C/C;     // der(u_C)
}

// Runge-Kutta-4 step
void RK4step(double t, double h, double *y);

// Experiment control
void Run()
{
    // Print model and simulation parameters:
    printf("# Rz=%g\n", Rz);
    printf("# C=%g\n", C);
    const double T    = 1/Freq;     // period of input signal
    const double TEND = 3*T;        // how many periods
    const double stepsize = T/400;  // RK4 step size (should be small)
    double y[SIZE] = { 0, };        // initial values
    double t = 0;                   // start time
    printf("#%8.6s  %8s  %8s  %8s  %8s\n", "t", "uin", "uout", "Id", "iout");
    while (t < TEND) {
        printf(" %8.6f  %8g  %8g  %8g  %8g\n", t, uin(t), uout(t,y), Id, iout(t,y));
        RK4step(t, stepsize, y);    // compute numerical method step
        t += stepsize;              // advance the simulation time
    }
    printf("\n");
}

int main()
{
    printf("# Simple rectifier simulation output\n");
    // 3 experiments, we use various load resistance Rz
    Rz =  50; Run();
    Rz = 100; Run();
    Rz = 200; Run();
    return 0;
}

//////////////////////////////////////////////////////////////////
// Runge-Kutta 4-th order method step
// Parameters: t = start time,
//             h = step size,
//             y = state vector
// Output:  new state in y
// Warning: it does not update time
void RK4step(double t, double h, double *y)
{
    double yin[SIZE];           // integrator input = derivative
    double ystart[SIZE];        // initial state
    // 4 stages results:
    double k1[SIZE];
    double k2[SIZE];
    double k3[SIZE];
    double k4[SIZE];

    int i;
    for (i = 0; i < SIZE; i++)  // save initial value
        ystart[i] = y[i];

    f(t, y, yin);               // yin = f(t, y(t))

    for (i = 0; i < SIZE; i++) {
        k1[i] = h * yin[i];     // k1 = h * f(t, y(t))
        y[i] = ystart[i] + k1[i] / 2;
    }

    f(t + h / 2, y, yin);       // yin = f(t+h/2, y(t)+k1/2)

    for (i = 0; i < SIZE; i++) {
        k2[i] = h * yin[i];     // k2 = h * f(t+h/2, y(t)+k1/2)
        y[i] = ystart[i] + k2[i] / 2;
    }

    f(t + h / 2, y, yin);       // yin = f(t+h/2, y(t)+k2/2)

    for (i = 0; i < SIZE; i++) {
        k3[i] = h * yin[i];     // k3 = h * f(t+h/2, y(t)+k2/2)
        y[i] = ystart[i] + k3[i];
    }

    f(t + h, y, yin);           // yin = f(t+h, y(t)+k3)

    for (i = 0; i < SIZE; i++) {
        k4[i] = h * yin[i];     // k4 = h * f(t+h, y(t)+k3)
        // Result:
        // y(t+h) = y(t) + k1/6 + k2/3 + k3/3 + k4/6;
        y[i] = ystart[i] + k1[i] / 6 + k2[i] / 3 + k3[i] / 3 + k4[i] / 6;
    }
    // Return: y = new state at time t+h
}
